#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from bibtexparser.middlewares.names import NameParts
from bibtexparser import model

# ##-- end 3rd party imports

from .name_parts import NameParts_d

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Either, Result
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type String  = model.String
    type StrLike = str|NameParts|list|set|String
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging
OBRACE : Final[str] = "{"
CBRACE : Final[str] = "}"

##--|

class StringTransform_m:
    """ Mixin for handling transform of strings
    refactored from bibtexparser middlewares.

    Implement _transform_raw_str,
    and call transform_string_like
    """

    def _transform_all_strings(self, vals:list[str]) -> Result[list[str], ValueError]:
        """Called for every python (value, not key) string found on Entry and String blocks
        Errors are modified in place.
        """
        res  = []
        errs = []
        for s in vals:
            match self._transform_raw_str(s):
                case ValueError() as x:
                    errs += x.args
                case str() as x:
                    res.append(x)
        else:
            if bool(errs):
                return ValueError(*errs)

            return res

    def _transform_nameparts(self, parts:NameParts) -> Result[NameParts, ValueError]:
        parts.first = self._transform_all_strings(parts.first)
        parts.last  = self._transform_all_strings(parts.last)
        parts.von   = self._transform_all_strings(parts.von)
        parts.jr    = self._transform_all_strings(parts.jr)
        return parts

    def transform_strlike(self, slike:StrLike) -> Result[StrLike,ValueError]:
        """
        Transform str likes: str,s
        """
        match slike:
            case model.String(value=str() as sval):
                match self.transform_strlike(sval):
                    case ValueError() as err:
                        res = err
                    case str() as x:
                        slike.value = x
                        res = slike
            case str() as sval if sval.startswith(OBRACE) and x.endswith(CBRACE):
                match self._transform_raw_str(sval[0:-1]):
                    case ValueError() as err:
                        res = err
                    case str() as x:
                        res = "".join([OBRACE, x, CBRACE])
            case str() as sval:
                res = self._transform_raw_str(sval)
            case NameParts() as np:
                res = self._transform_nameparts(np)
            case list() as vals:
                res = []
                errs = []
                for x in vals:
                    match self.transform_strlike(x):
                        case ValueError() as err:
                            errs += err.args
                        case str() as x:
                            res.append(x)
                else:
                    if bool(errs):
                        res = ValueError(*errs)
            case set() as vals:
                res  = []
                errs = []
                for x in vals:
                    match self.transform_strlike(x):
                        case ValueError() as err:
                            errs += err.args
                        case str() as x:
                            res.append(x)
                else:
                    if bool(errs):
                        res = ValueError(*errs)
                    else:
                        res = set(res)
            case x:
                logging.info(
                    f" [{self.metadata_key()}] Cannot python-str transform: {x}"
                    f" with value type {type(x)}"
                )
                res = x

        return res



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

import jgdv
from jgdv import Proto, Mixin
from bibble import _interface as API  # noqa: N812
from bibble.util.middlecore import IdenBidiMiddleware
from bibtexparser import model

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from bibtexparser import Library
    type Entry  = model.Entry
    type String = model.String
    type Field  = model.Field
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
OBRACE : Final[str]           = "{"
CBRACE : Final[str]           = "}"
DQUOTE : Final[str]           = '"'
STRINGS_CAN_BE_UNESCAPED_INTS = False
ENTRY_POTENTIALLY_INT_FIELDS  = [
    "year",
    "month",
    "volume",
    "number",
    "pages",
    "edition",
    "chapter",
    "issue",
]
# Body:

@Proto(API.BidirectionalMiddleware_p)
class BraceWrapper(IdenBidiMiddleware):
    """ a Bidirectional replacement for bibtexparsers 'Add/RemoveEnclosingMiddleware' """

    def __init__(self, *, wrappers:Maybe[tuple[str,str]]=None, on_ints:bool=True, **kwargs):
        super().__init__(**kwargs)
        self._wrapper = wrappers or (OBRACE, CBRACE)
        self._on_ints = on_ints

    def _wrap(self, value: str, *, maybe_int_rule:bool=False) -> str:
        """ Take 'value' and wrap it in (by default) braces """
        if maybe_int_rule and not self._on_ints and value.isdigit():
            return value

        ochar, cchar = self._wrapper
        return f"{ochar}{value}{cchar}"

    def _unwrap(self, value: str) -> str:
        """ remove wrapping (by default) braces """
        ochar, cchar = self._wrapper
        start, end   = len(ochar), 0 - len(cchar)
        value        = value.strip()
        if value.startswith(ochar) and value.endswith(cchar):
            return value[start:end]
        else:
            return value

    def write_transform_Entry(self, entry:Entry, *args, **kwargs) -> list[Entry]:
        for field in entry.fields:
            field.value = self._wrap(field.value, maybe_int_rule=field.key in ENTRY_POTENTIALLY_INT_FIELDS)
        else:
            return [entry]

    def write_transform_String(self, string:String, *args, **kwargs) -> list[String]:
        string.value = self._wrap(string.value, maybe_int_rule=STRINGS_CAN_BE_UNESCAPED_INTS)
        return [string]

    def read_transform_Entry(self, entry: Entry, library: Library) -> list[Entry]:
        for field in entry.fields:
            field.value = self._unwrap(field.value)
        else:
            return [entry]

        Never()

    def read_transform_String(self, string: String, library: Library) -> list[String]:
        string.value = self._unwrap(string.value)
        return [string]



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest
from bibtexparser import model
import bibble._interface as API
from bibble.fields import FieldSorter

logging = logmod.root

class TestFieldSorter:

    @pytest.fixture(scope="function")
    def sorter(self):
        return FieldSorter(first=["title", "author", "editor"], last=["doi", "url", "file"])

    @pytest.fixture(scope="function")
    def entry(self):
        return model.Entry("book", "blah", [])

    def get_field_order(self, entry) -> list[str]:
        return [x.key for x in entry.fields]

    ##--|

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self, sorter):
        match sorter:
            case API.Middleware_p():
                assert(True)
            case x:
                 assert(False), x

    def test_basic(self, sorter, entry):
        result = sorter.transform_Entry(entry, None)
        assert(result == [entry])

    def test_one_field(self, sorter, entry):
        entry.set_field(model.Field("test", ""))
        result = sorter.transform_Entry(entry, None)
        assert(result == [entry])

    def test_two_fields(self, sorter, entry):
        entry.set_field(model.Field("author", ""))
        entry.set_field(model.Field("title", ""))
        assert(self.get_field_order(entry) == ["author", "title"])
        result = sorter.transform_Entry(entry, None)
        assert(self.get_field_order(entry) == ["title", "author"])

    def test_firsts_and_lasts(self, sorter, entry):
        entry.set_field(model.Field("author", ""))
        entry.set_field(model.Field("file", ""))
        entry.set_field(model.Field("title", ""))
        entry.set_field(model.Field("misc", ""))
        assert(self.get_field_order(entry) == ["author", "file", "title", "misc"])
        result = sorter.transform_Entry(entry, None)
        assert(self.get_field_order(entry) == ["title", "author", "misc", "file"])

    def test_key_stemming(self, sorter, entry):
        entry.set_field(model.Field("file2", ""))
        entry.set_field(model.Field("file3", ""))
        entry.set_field(model.Field("file4", ""))
        entry.set_field(model.Field("file", ""))
        assert(self.get_field_order(entry) == ["file2", "file3", "file4", "file"])
        result = sorter.transform_Entry(entry, None)
        assert(self.get_field_order(entry) == ["file", "file2", "file3", "file4"])

    def test_stem_sorting(self, sorter, entry):
        entry.set_field(model.Field("file2", ""))
        entry.set_field(model.Field("url", ""))
        entry.set_field(model.Field("doi", ""))
        entry.set_field(model.Field("file", ""))
        assert(self.get_field_order(entry) == ["file2", "url", "doi", "file"])
        result = sorter.transform_Entry(entry, None)
        assert(self.get_field_order(entry) == ["doi", "url", "file", "file2"])



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Mixin
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from jgdv.files.tags import SubstitutionFile

# ##-- end 3rd party imports

# ##-- 1st party imports
from bibble.util.mixins import ErrorRaiser_m, FieldMatcher_m
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Mixin(ErrorRaiser_m, FieldMatcher_m)
class FieldSubstitutor(IdenBlockMiddleware):
    """
      For a given field(s), and a given jgdv.SubstitutionFile,
    replace the field value as necessary in each entry.

    If force_single_value is True, only the first replacement will be used,
    others will be discarded

    eg: for target=['tags'],
    and subs({'AI': ['artificial_intelligence', 'agents', 'machine_learning'])
    and entry(fields={'tags': 'ai'})
    will give: entry(fields={''tags': ['artificial_intelligence', 'agents', 'machine_learning']})

    """

    def __init__(self, *, fields:list[str], subs:SubstitutionFile, force_single_value:bool=False, **kwargs):
        super().__init__(**kwargs)
        match fields:
            case list():
                self._target_fields = fields
            case x:
                raise TypeError(type(x))

        self._subs               = subs
        self._force_single_value = force_single_value
        self.set_field_matchers(white=self._target_fields, black=[])

    def transform_Entry(self, entry, library):
        if self._subs is None or not bool(self._subs):
            return [entry]

        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x))

    def field_h(self, field, entry):
        """  """
        match field.value:
            case str() as value if self._force_single_value:
                head, *_ = list(self._subs.sub(value))
                return [model.Field(field.key, head)]
            case str() as value:
                subs = list(self._subs.sub(value))
                return [model.Field(field.key, subs)]
            case list() | set() as value:
                result = self._subs.sub_many(*value)
                return [model.Field(field.key, result)]
            case value:
                return ValueError("Unsupported replacement field value type", entry.key, type(value))

        return []



"""
bibble.bidi : A module of bidirectional middlewares.

Most are simple wrappers around the paired unidirectional middlewares.
But some, like BraceWrapper, is purely bidirection

"""
from .bidi_paths    import BidiPaths
from .bidi_isbn     import BidiIsbn
from .bidi_latex    import BidiLatex
from .brace_wrapper import BraceWrapper
from .bidi_names    import BidiNames
from .bidi_tags     import BidiTags



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from bibtexparser import model
from jgdv import Maybe, Proto

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from bibtexparser.libary import Library
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.CustomWriteBlock_p)
class MetaBlock(model.Block):
    """ A Metadata Block baseclass that does not get written out (typically),
    But can hold information about the library
    """

    @classmethod
    def find_in(cls, lib:Library) -> Maybe[Self]:
        """ Find a block of this cls in a given library """
        for block in lib.blocks:
            if isinstance(block, cls):
                return block
        else:
            return None

    def __init__(self, **kwargs):
        super().__init__(0)
        self.data = dict(kwargs)

    def visit(self, *args, **kwargs) -> list[str]:
        return []


class FailedBlock(model.MiddlewareErrorBlock):
    """ Records errors encountered by a middleware """

    def __init__(self, *, block:model.Block, error:Exception, source:type|str):
        super().__init__(block, error)
        self._block_type = type(block).__name__
        match source:
            case API.Middleware_p() as mw:
                self.source_middleware = type(source).__name__
            case type():
                self.source_middleware = source.__name__
            case str():
                self.source_middleware = source
            case x:
                raise TypeError(type(x))


    def __repr__(self):
        key = self.ignore_error_block.key
        return f"<{self.__class__.__name__}: {key}>"

    def report(self, *, i:int, total:int, source_file:Maybe[str|pl.Path]=None, **kwargs) -> list[str]:
        match source_file:
            case None:
                report = f"({i}/{total}) [{self.source_middleware}] Bad <{self._block_type}>: {self.start_line} : {self.error}"
            case str() | pl.Path():
                report = f"({i}/{total}) [{self.source_middleware}] Bad <{self._block_type}>: {source_file}:{self.start_line} : {self.error}"
            case x:
                raise TypeError(type(x))

        return [report]




#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from bibtexparser.library import Library
from bibtexparser import model
import bibble._interface as API
from .. import FieldAccumulator
from .. import _interface as API_F

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestFieldAccumulator:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match FieldAccumulator(name="test", fields=[]):
            case API.AdaptiveMiddleware_p():
                assert(True)
            case x:
                 assert(False), x


    def test_accumulation(self):
        mid = FieldAccumulator(name="test", fields=["author"])
        entry1 = model.Entry("test", "first",  [model.Field("author", "bob")])
        entry2 = model.Entry("test", "second", [model.Field("author", "jill")])
        entry3 = model.Entry("test", "third",  [model.Field("author", "bob")])
        lib = Library()
        lib.add(entry1)
        lib.add(entry2)
        lib.add(entry2)
        assert(mid.transform(lib) is lib)
        match API_F.AccumulationBlock.find_in(lib):
            case API_F.AccumulationBlock() as bl:
                assert(bl.collection == {"bob", "jill"})
            case x:
                 assert(False), x


    def test_multi_field_accumulation(self):
        mid = FieldAccumulator(name="test", fields=["author", "editor"])
        entry1 = model.Entry("test", "first",  [model.Field("author", "bob")])
        entry2 = model.Entry("test", "second", [model.Field("author", "jill")])
        # author -> editor here:
        entry3 = model.Entry("test", "third",  [model.Field("editor", "bob")])
        lib = Library()
        lib.add(entry1)
        lib.add(entry2)
        lib.add(entry2)
        assert(mid.transform(lib) is lib)
        match API_F.AccumulationBlock.find_in(lib):
            case API_F.AccumulationBlock() as bl:
                assert(bl.collection == {"bob", "jill"})
            case x:
                 assert(False), x
            




#!/usr/bin/env python3
"""

"""
# mypy: disable-error-code="attr-defined"

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from copy import deepcopy
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import jgdv
from jgdv import Proto, Mixin
from jgdv.debugging.timeblock_ctx import TimeBlock_ctx
from bibtexparser import model
from bibtexparser.model import MiddlewareErrorBlock

# ##-- end 3rd party imports

# ##-- 1st party imports
from bibble import _interface as API
from . import _interface as API_W
from bibble.util.mixins import MiddlewareValidator_m
from bibble.model import MetaBlock
from bibble.util import PairStack

from ._util import Runner_m
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from bibtexparser.library import Library
    from bibtexparser.writer import BibtexFormat

    type Middleware = API.Middleware_p | API.BidirectionalMiddleware_p
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class _VisitEntry_m:

    def make_header(self, library, file:None|pl.Path=None) -> list[str]:
        return []

    def make_footer(self, library, file:None|pl.Path=None) -> list[str]:
        return []

    def visit(self, block) -> list[str]:
        match block:
            case x if isinstance(x, API.CustomWriteBlock_p):
                assert(hasattr(x, "visit"))
                return x.visit(self)
            case MetaBlock():
                return []
            case model.Entry():
                return self.visit_entry(block)
            case model.String():
                return self.visit_string(block)
            case model.Preamble():
                return self.visit_preamble(block)
            case model.ExplicitComment():
                return self.visit_expl_comment(block)
            case model.ImplicitComment():
                return self.visit_impl_comment(block)
            case model.ParsingFailedBlock():
                return self.visit_failed_block(block)
            case MiddlewareErrorBlock():
                return []
            case _:
                raise ValueError(f"Unknown block type: {type(block)}")

class _Visitors_m:

    def visit_entry(self, block:model.Entry) -> list[str]:
        res = ["@", block.entry_type, "{", block.key, ",\n"]
        field: model.Field
        for i, field in enumerate(block.fields):
            res.append(self._align_key(field.key))
            res.append(field.value)
            if self.format.trailing_comma or i < len(block.fields) - 1:
                res.append(",")
                res.append("\n")
        else:
            res.append("}\n")
            return res

    def visit_string(self, block:model.String) -> list[str]:
        return [
            "@string{",
            block.key,
            self._value_sep,
            block.value,
            "}\n",
        ]

    def visit_preamble(self, block:model.Preamble) -> list[str]:
        return [f"@preamble{{{block.value}}}\n"]

    def visit_impl_comment(self, block:model.ImplicitComment) -> list[str]:
        # Note: No explicit escaping is done here - that should be done in middleware
        return [block.comment, "\n"]

    def visit_expl_comment(self, block:model.ExplicitComment) -> list[str]:
        return ["@comment{", block.comment, "}\n"]

    def visit_failed_block(self, block:model.ParsingFailedBlock) -> list[str]:
        format_line            = self.format.parsing_failed_comment
        lines                  = len(block.raw.splitlines())
        err                    = f"<{block.error.__class__.__name__}> : {block.error}"
        parsing_failed_comment = format_line.format(n=lines, err=err)
        return [parsing_failed_comment, "\n",
                block.raw, "\n",
                API_W.FAIL_END, "\n"]
    ##--|

@Proto(jgdv.protos.Visitor_p, API.Writer_p)
@Mixin(_VisitEntry_m, _Visitors_m, Runner_m, MiddlewareValidator_m)
class BibbleWriter:
    """ A Refactored bibtexparser writer
    Uses visitor pattern

    Note: visit method are responsible for new lines
    """
    _value_sep             : str
    _value_column          : Maybe[int]
    _middlewares           : list[Middleware]
    format                 : BibtexFormat

    def __init__(self, stack:PairStack|list[Middleware], *, format:Maybe[BibtexFormat]=None, logger:Maybe[Logger]=None):
        self._value_sep    = API_W.VAL_SEP
        self._value_column = None
        self._logger       = logger or logging
        match stack:
            case PairStack():
                self._middlewares = stack.write_stack()
            case list():
                self._middlewares = stack
            case x:
                raise TypeError(type(x))

        match format:
            case None:
                self.format = deepcopy(API_W.default_format())
            case BibtexFormat():
                self.format = deepcopy(format)
            case x:
                raise TypeError(type(x))

        self.exclude_middlewares(API.ReadTime_p)

    def write(self, library, *, file:None|pl.Path=None, append:Maybe[list[Middleware]]=None) -> str:
        """ Write the library to a string, and possbly a file
        # TODO write failure reports to a separate file
        """

        self._calculate_auto_value_align(library)

        with TimeBlock_ctx(logger=logging,
                           enter="--> Write Transforms: Start",
                           exit="<-- Write Transforms:",
                           level=logmod.INFO) as ctx:
            transformed = self._run_writewares(library, append=append)

        string_pieces : list[str] = []
        string_pieces += self.make_header(transformed, file)
        for i, block in enumerate(transformed.blocks):
            # Get string representation (as list of strings) of block
            string_block_pieces = self.visit(block)
            string_pieces.extend(string_block_pieces)
            # Separate Blocks
            if i < len(transformed.blocks) - 1:
                string_pieces.append(self.format.block_separator)
        else:
            string_pieces.extend(self.make_footer(transformed, file))
            result : str = "".join(str(x) for x in string_pieces)

        # Reset the value column:
        self._value_column = None
        match file:
            case pl.Path():
                file.write_text(result)
                return result
            case _:
                return result

    def _calculate_auto_value_align(self, library: Library) -> None:
        """
        Sets the separation between keys and the value separator.
        If its already set, does nothing.
        If the format specifies a value, uses that.
        Otherwise calulates it from the larges field key
        """
        if self._value_column is not None:
            return

        match self.format.value_column:
            case int() as x:
                self._value_column = x
            case _:
                max_key_len = 0
                for entry in library.entries:
                    for key in entry.fields_dict:
                        max_key_len = max(max_key_len, len(key))
                    ##--|
                else:
                    self._value_column = max_key_len + len(self._value_sep)

    def _align_key(self, key: str) -> str:
        """ take {key} and make {key}{padding}{sep},
        Padding is from '_calculate_auto_value_align', the largest key length.
        Sep is typically '='.

        eg: _align_key('blah') -> 'blah   = '
        """
        match (self.format.value_column - len(key) - len(self._value_sep)):
            case x if 0 <= x:
                return f"{self.format.indent}{key}{' '*x}{self._value_sep}"
            case x:
                return f"{self.format.indent}{key}{' '*x}{self._value_sep}"




#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Proto, Mixin
import bibtexparser
import bibtexparser.model as model
from bibtexparser.library import Library
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware, LibraryMiddleware)

# ##-- end 3rd party imports

import bibble._interface as API
from . import _interface as MAPI
from bibble.util.middlecore import IdenLibraryMiddleware

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--|

class DuplicateFinder(IdenLibraryMiddleware):

    def transform(self, library:Library) -> Library:
        keys = set()
        for entry in library.entries:
            if entry.key in keys:
                # duplicate
                pass
            else:
                keys.add(entry.key)
        else:
            return library


class DuplicateKeyHandler(IdenLibraryMiddleware):
    """ take duplicate entries and edit their key to be unique """

    def transform(self, library:Library):
        if not bool(library.failed_blocks):
            return library

        key_count, field_count = 0, 0
        self.logger().info("Handling %s failed blocks", len(library.failed_blocks))
        for failed in library.failed_blocks:
            match failed:
                case model.DuplicateBlockKeyBlock():
                    self._dedup_key(failed, library)
                    key_count   += 1
                case model.DuplicateFieldKeyBlock():
                    self._dedup_fields(failed, library)
                    field_count += 1
                case _:
                    self.logger().info("Skipping block: %s", failed)
                    pass
        else:
            self.logger().info("Adjusted %s duplicate keys ", key_count)
            self.logger().info("Adjusted %s duplicate fields ", field_count)
            return library

    def _dedup_key(self, failed, library) -> None:
        uuid          = uuid1().hex
        duplicate     = failed.ignore_error_block
        original      = duplicate.key
        duplicate.key = f"{duplicate.key}_dup_{uuid}"
        library.add(duplicate)
        library.remove(failed)
        self.logger().warning("Duplicate Key found: %s -> %s", original, duplicate.key)

    def _dedup_fields(self, failed, library) -> None:
        entry       = failed.ignore_error_block
        found       = set()
        duplicates  = set()
        for field in entry.fields:
            if field.key not in found:
                found.add(field.key)
                continue

            duplicates.add(field.key)
            count = 2

            while (curr:=f"{field.key}_{count}") in found:
                duplicates.add(field.key)
                count += 1
                if 100 < count:
                    raise ValueError("Deduplicating fields is stuck")
            else:
                field.key = curr
                found.add(curr)
        else:
            self.logger().warning("Duplicate Fields (%s): %s",
                                  entry.key,
                                  duplicates)
            library.add(entry)
            library.remove(failed)



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from jgdv import Mixin, Proto

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from bibble.util.mixins import FieldMatcher_m
from bibble.util.middlecore import IdenBlockMiddleware
from bibble.util.name_parts import NameParts_d

# ##-- end 1st party imports

import pyparsing as pp
from . import _interface as API_N

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Result
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Block = model.Block
    type Field = model.Field
    type Entry = model.Entry
    from bibtexparser.library import Library
    type Parser = pp.core.ParserElement
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

PARSE_STATE = API_N.NameSplitState_e
##--|

class _SplitAuthors_m:
    """ Adapated from bibtexparser's split_multiple_persons_names, originally by Blair Bonnett

    Splits names by intermediary 'and's.
    Like its original, treats non-breaking space and '~'s as regular chars not whitespace.

    'and's within braces are returned un modified.
    eg: '{Simon and Schuster}' -> ['{Simon and Schuster}']

    """

    def _build_split_parser(self) -> Parser:
        return pp.Literal("and")

    def _split_authors(self, val:str, *, strict=True) -> list[str]:
        return self._split_authors_fsm(val, strict=strict)

    def _split_authors_pp(self, val:str, *, strict=True) -> list[str]:
        """
        TODO
        """
        val = val.strip()
        if not bool(val):
            return []

        parser = self._build_split_parser()

        return []

    def _split_authors_fsm(self, val:str, *, strict=True) -> list[str]:
        val = val.strip()
        if not bool(val):
            return []

        # Processing variables.
        step         = PARSE_STATE.start_ws  # Current step.
        pos          = 0  # Current position in string.
        bracelevel   = 0  # Current bracelevel.
        spans        = [[0]]  # Spans of names within the string.
        possible_end = 0  # Possible end position of a name.
        whitespace   = API_N.NAME_WHITESPACE # Allowed whitespace characters.

        # Loop over the string.
        namesiter = iter(val)
        for char in namesiter:
            pos += 1
            match char:
                case "\\":
                    try:
                        next(namesiter)
                    except StopIteration:
                        # If we're at the end of the string, then the \ is just a \.
                        pass
                    pos += 1
                    continue
                case API_N.OBRACE:
                    # Change in brace level.
                    if step == API_N.NameSplitState_e.next_word:
                        spans[-1].append(possible_end)
                        spans.append([pos - 1])

                    bracelevel += 1
                    step = API_N.NameSplitState_e.start_ws
                    continue
                case API_N.CBRACE:
                    if bracelevel:
                        bracelevel -= 1

                    step = API_N.NameSplitState_e.start_ws
                    continue
                case _:
                    pass

            # Ignore everything inside a brace.
            if bracelevel:
                step = API_N.NameSplitState_e.start_ws
                continue

            match step:
                case API_N.NameSplitState_e.start_ws if char in whitespace:
                    # Looking for a whitespace character to start the ' and '. When we find
                    # one, mark it as the possible end of the previous word.
                    step = API_N.NameSplitState_e.find_a
                    possible_end = pos - 1
                case API_N.NameSplitState_e.find_a if char in ("a", "A"):
                    # Looking for the letter "a".
                    step = API_N.NameSplitState_e.find_n
                case API_N.NameSplitState_e.find_a if char not in whitespace:
                    # NB, we can have multiple whitespace characters so we need to handle that here.
                    step = API_N.NameSplitState_e.start_ws
                case API_N.NameSplitState_e.find_n if char in ("n", "N"):
                    # Looking for the letter n.
                    step = API_N.NameSplitState_e.find_d
                case API_N.NameSplitState_e.find_n if char in whitespace:
                    step = API_N.NameSplitState_e.find_a
                    possible_end = pos - 1
                case API_N.NameSplitState_e.find_n:
                    step = API_N.NameSplitState_e.start_ws
                case API_N.NameSplitState_e.find_d if char in ("d", "D"):
                    # Looking for the letter d.
                    step = API_N.NameSplitState_e.end_ws
                case API_N.NameSplitState_e.find_d if char in whitespace:
                    step = API_N.NameSplitState_e.find_a
                    possible_end = pos - 1
                case API_N.NameSplitState_e.find_d:
                    step = API_N.NameSplitState_e.start_ws
                case API_N.NameSplitState_e.end_ws if char in whitespace:
                    # And now the whitespace to end the ' and '.
                    step = API_N.NameSplitState_e.next_word
                case API_N.NameSplitState_e.end_ws:
                    step = API_N.NameSplitState_e.start_ws
                case API_N.NameSplitState_e.next_word if char not in whitespace:
                    # Again, we need to handle multiple whitespace characters. Keep going
                    # until we find the start of the next word.
                    # Finish the previous word span, start the next,
                    # and do it all again.
                    spans[-1].append(possible_end)
                    spans.append([pos - 1])
                    step = API_N.NameSplitState_e.start_ws

        # Finish the last word.

        spans[-1].append(None)

        # Extract and return the names.
        return [val[start:end] for start, end in spans]

class _NameToParts_m:
    """ Adapted from bibtexparser's parse_single_name_into_parts, originally by Blair Bonnett.

    Parses an individual name into a NameParts_d, a simple data structure containing:
    - first : list. First names.
    - von   : list.
    - last  : list. Last Names.
    - jr    : list.

    Bibtex Names are of one of the forms:
    - first von last
    - von last, first
    - von last, jr, first
    """

    def _build_parts_parser(self) -> Parser:
        return pp.Literal("and")

    def _name_to_parts(self, val:str, *, strict=True) -> NameParts_d:
        val = val.strip()
        if not bool(val):
            return NameParts_d()

        sections, cases = self._parse_name_fsm(val, strict=strict)
        # No non-whitespace input.
        if not sections or not any(bool(section) for section in sections):
            return NameParts_d()

        match sections:
            case [x]:
                parts = self._first_von_last(x, cases)
            case [*xs]:
                parts = self._von_last_first(xs, cases)

        return parts

    def _parse_name_pp(self, val, *, strict=True) -> tuple[list, list]:
        """ TODO """
        return [], []

    def _parse_name_fsm(self, val, *, strict=True) -> tuple[list, list]:
        # We'll iterate over the input once, dividing it into a list of words for
        # each comma-separated section. We'll also calculate the case of each word
        # as we work.
        sections    = [[]]   # Sections of the name.
        cases       = [[]]   # 1 = uppercase, 0 = lowercase, -1 = caseless.
        word        = []     # Current word.
        case        = -1     # Case of the current word.
        level       = 0      # Current brace level.
        bracestart  = False  # Will the next character be the first within a brace?
        controlseq  = True   # Are we currently processing a control sequence?
        specialchar = None   # Are we currently processing a special character?
        whitespace  = API_N.NAME_WHITESPACE

        # Using an iterator allows us to deal with escapes in a simple manner.
        nameiter = iter(val)
        for char in nameiter:
            # An escape.
            match char:
                case "\\":
                    try:
                        escaped = next(nameiter)
                        # BibTeX doesn't allow whitespace escaping. Copy the slash and fall
                        # through to the normal case to handle the whitespace.
                        if escaped in whitespace:
                            word.append(char)
                            char = escaped
                        else:
                            if bracestart:
                                # Is this the first character in a brace?
                                bracestart = False
                                controlseq = escaped.isalpha()
                                specialchar = True
                                # Can we use it to determine the case?
                            elif (case == -1) and escaped.isalpha():
                                if escaped.isupper():
                                    case = 1
                                else:
                                    case = 0

                            # Copy the escape to the current word and go to the next
                            # character in the input.
                            word.append(char)
                            word.append(escaped)
                            continue
                    except StopIteration:
                        # If we're at the end of the string, then the \ is just a \.
                        word.append(char)
                case API_N.OBRACE:
                    # Start of a braced expression.
                    level += 1
                    word.append(char)
                    bracestart = True
                    controlseq = False
                    specialchar = False
                    continue
                case API_N .CBRACE:
                    # All the below cases imply this (and don't test its previous value).
                    bracestart = False
                    # End of a braced expression.
                    # Check and reduce the level.
                    if level:
                        level -= 1
                    else:
                        if strict:
                            raise ValueError(name=val, reason="Unmatched closing brace")
                        word.insert(0, "{")

                    # Update the state, append the character, and move on.
                    controlseq = False
                    specialchar = False
                    word.append(char)
                    continue
                case _ if level:
                    # All the below cases imply this (and don't test its previous value).
                    bracestart = False
                    # Inside a braced expression.
                    # Is this the end of a control sequence?
                    if controlseq:
                        if not char.isalpha():
                            controlseq = False
                    # If it's a special character, can we use it for a case?
                    elif specialchar:
                        if (case == -1) and char.isalpha():
                            if char.isupper():
                                case = 1
                            else:
                                case = 0

                    # Append the character and move on.
                    word.append(char)
                    continue

            # End of a word.
            # NB. we know we're not in a brace here due to the previous case.
                case x if x == "," or x in whitespace:
                    # All the below cases imply this (and don't test its previous value).
                    bracestart = False
                    # Don't add empty words due to repeated whitespace.
                    if word:
                        sections[-1].append("".join(word))
                        word = []
                        cases[-1].append(case)
                        case = -1
                        controlseq = False
                        specialchar = False

                    # End of a section.
                    if char == ",":
                        if len(sections) < 3:
                            sections.append([])
                            cases.append([])
                        elif strict:
                            raise ValueError(name=val, reason="Too many commas")
                    continue
            ##--|
            # Regular character.
            word.append(char)
            if (case == -1) and char.isalpha():
                if char.isupper():
                    case = 1
                else:
                    case = 0
        else:
            pass
        ##--|
        # Unterminated brace?
        if level:
            if strict:
                raise ValueError(name=val, reason="Unterminated opening brace")
            while level:
                word.append(API_N.CBRACE)
                level -= 1

        # Handle the final word.
        if word:
            sections[-1].append("".join(word))
            cases[-1].append(case)

        # Get rid of trailing sections.
        if not sections[-1]:
            # Trailing comma?
            if (len(sections) > 1) and strict:
                raise ValueError(name=val, reason="Trailing comma at end of name")
            sections.pop(-1)
            cases.pop(-1)

        return sections, cases

    def _first_von_last(self, p0, cases) -> NameParts_d:
        # Form 2: "First von Last"
        parts = NameParts_d()

        # One word only: last cannot be empty.
        if len(p0) == 1:
            parts.last = p0

        # Two words: must be first and last.
        elif len(p0) == 2:
            parts.first = p0[:1]
            parts.last = p0[1:]

        # Need to use the cases to figure it out.
        else:
            cases = cases[0]

            # - First is the longest sequence of words starting with uppercase
            # that is not the whole string.
            # - von is then the longest sequence # whose last word starts with
            # lowercase that is not the whole # string.
            # - Last is the rest.
            # NB., this means last cannot be empty.

            # At least one lowercase letter.
            if 0 in cases:
                # Index from end of list of first and last lowercase word.
                firstl = cases.index(0) - len(cases)
                lastl = -cases[::-1].index(0) - 1
                if lastl == -1:
                    lastl -= 1  # Cannot consume the rest of the string.

                # Pull the parts out.
                parts.first = p0[:firstl]
                parts.von = p0[firstl : lastl + 1]
                parts.last = p0[lastl + 1 :]

            # No lowercase: last is the last word, first is everything else.
            else:
                parts.first = p0[:-1]
                parts.last = p0[-1:]
        ##--|
        return parts

    def _von_last_first(self, sections, cases) -> NameParts_d:
        # Form 2 ("von Last, First") or 3 ("von Last, jr, First")
        # As long as there is content in the first name partition, use it as-is.
        parts = NameParts_d()
        first = sections[-1]
        if first and first[0]:
            parts.first = first

        # And again with the jr part.
        if len(sections) == 3:
            jr = sections[-2]
            if jr and jr[0]:
                parts.jr = jr

        # Last name cannot be empty; if there is only one word in the first
        # partition, we have to use it for the last name.
        last = sections[0]
        if len(last) == 1:
            parts.last = last
            return parts

        # Have to look at the cases to figure it out.
        lcases = cases[0]

        def rindex(k, x, default):
            """Returns the index of the rightmost occurrence of x in k."""
            for i in range(len(k) - 1, -1, -1):
                if k[i] == x:
                    return i
            return default

        # Check if at least one of the words is lowercase
        if 0 in lcases:
            # Excluding the last word, find the index of the last lower word
            split = rindex(lcases[:-1], 0, -1) + 1
            parts.von = sections[0][:split]
            parts.last = sections[0][split:]

        # All uppercase => all last.
        else:
            parts.last = sections[0]

        ##--|
        return parts

##--|

@Proto(API.ReadTime_p)
@Mixin(FieldMatcher_m, _SplitAuthors_m, _NameToParts_m)
class NameReader(IdenBlockMiddleware):
    """ A Refactored version of bibtexparser's SplitNameParts and SeparateCoAuthors
    """
    _whitelist = ("author", "editor", "translator")

    def __init__(self, *, parts:bool=True, authors:bool=True,  **kwargs):
        super().__init__(**kwargs)
        self._do_split_authors = authors
        self._do_name_parts = parts
        self.set_field_matchers(white=self._whitelist, black=[])
        if self._do_name_parts and not self._do_split_authors:
            raise ValueError("Can't generate name parts if you don't split authors")

    def on_read(self):
        Never()

    def transform_Entry(self, entry:Entry, library:Library) -> list[Entry]:
        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x))

    def field_h(self, field:Field, entry:Entry) -> Result[list[Field], Exception]:
        result = []
        match self._do_split_authors:
            case True:
                authors = self._split_authors(field.value)
            case False:
                authors = field.value
            case x:
                raise TypeError(type(x))

        match authors:
            case str():
                pass
            case [*xs] if self._do_name_parts:
                parts = [self._name_to_parts(x) for x in xs]
                result.append(model.Field(field.key, parts))
            case [*xs]:
                result.append(model.Field(field.key, list(xs)))
            case x:
                raise TypeError(type(x))

        return result



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Proto, Mixin
import bibtexparser
import bibtexparser.model as model
from bibtexparser.library import Library
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)

# ##-- end 3rd party imports

# ##-- 1st party imports
from bibble import _interface as API
from . import _interface as API_F
from bibble.util.mixins import ErrorRaiser_m, FieldMatcher_m
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Entry = model.Entry

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Mixin(ErrorRaiser_m, FieldMatcher_m)
class CleanUrls(IdenBlockMiddleware):
    """ Strip unnecessary doi and dblp prefixes from urls  """

    _whitelist = (API_F.DOI_K, API_F.URL_K, API_F.EE_K)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_field_matchers(white=self._whitelist, black=[])

    def on_read(self):
        Never()

    def transform_Entry(self, entry:Entry, library:Library):
        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x))

    def field_h(self, field, entry):
        fields = []
        match field.value:
            case str() as value if value.startswith(API_F.DOI_PREFIX) and (field.key  == API_F.DOI_K or API_F.DOI_K not in entry):
                # Remove doi prefix
                clean = value.removeprefix(API_F.DOI_PREFIX)
                fields.append(model.Field(API_F.DOI_K, clean))
            case str() as value if value.startswith(API_F.BIBDB_PREFIX) and API_F.SOURCE_K not in entry:
                # cleanup dblp
                url = "".join([API_F.DBLP_PREFIX, value])
                fields.append(model.Field(API_F.BIBURL_K, url))
                fields.append(model.Field(API_F.SOURCE_K, "dblp computer science bibliography, https://dblp.org"))
            case str() as value if field.key == API_F.EE_K and API_F.URL_K not in entry:
                # ee -> url
                fields.append(model.Field(API_F.URL_K, value))
                fields.append(model.Field(API_F.EE_K, ""))
            case _:
                pass

        return fields

@Mixin(ErrorRaiser_m, FieldMatcher_m)
class ExpandUrls(IdenBlockMiddleware):
    """ TODO expand shortened urls """

    def on_read(self):
        Never()

    def transform_Entry(self, entry:Entry, library:Library):
        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x))

    def field_h(self, field, entry):
        raise NotImplementedError()



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

import bibble._interface as API
from bibtexparser import model, Library
from .. import ApplyMetadata, FileCheck, _interface as MAPI

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestApplyMetadata:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match ApplyMetadata():
            case API.AdaptiveMiddleware_p():
                assert(True)
            case x:
                 assert(False), x

    def test_no_file_entry(self):
        entry = model.Entry("test", "test:blah", [])
        lib   = Library([entry])
        mid   = ApplyMetadata()
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(not bool(lib.failed_blocks))
            case x:
                 assert(False), x


    def test_orphaned_entry(self):
        field = model.Field("file", "does/not/exist.pdf")
        entry = model.Entry("test", "test:blah", [field])
        lib   = Library([entry])
        mid   = ApplyMetadata()
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(not bool(lib.failed_blocks))
                assert(l2.entries[0].fields_dict.get(MAPI.ORPHANED_K, None) is not None)
            case x:
                 assert(False), x


    def test_noop_from_meta_match(self, mocker, tmpdir, caplog):
        logger = logmod.getLogger("test")
        caplog.set_level(logmod.INFO, logger.name)
        tmpfile = pl.Path(tmpdir) / "test.pdf"
        tmpfile.touch()
        assert(tmpfile.exists())
        field = model.Field("file", tmpfile)
        entry = model.Entry("test", "test:blah", [field])
        lib   = Library([entry])
        mid   = ApplyMetadata(logger=logger)
        mid.metadata_matches_entry = mocker.Mock(return_value=True)
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(not bool(lib.failed_blocks))
                assert("No Metadata Update Necessary" in caplog.text)
            case x:
                 assert(False), x
                 


    def test_locked_entry(self, mocker, tmpdir):
        tmpfile = pl.Path(tmpdir) / "test.pdf"
        tmpfile.touch()
        assert(tmpfile.exists())
        field                      = model.Field("file", tmpfile)
        entry                      = model.Entry("test", "test:blah", [field])
        lib                        = Library([entry])
        mid                        = ApplyMetadata()
        mid.pdf_is_modifiable      = mocker.Mock(return_value=False)
        mid.metadata_matches_entry = mocker.Mock(return_value=False)
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(bool(l2.failed_blocks))
                assert(l2.entries[0].fields_dict.get(MAPI.PDF_LOCKED_K, None) is not None)
            case x:
                assert(False), x

    @pytest.mark.skip
    def test_todo(self):
        """
        Still to test:
        - modifiable failure
        - original metadata backup
        - epub update
        - pdf update
        - pdf validation
        - pdf finalization
        - get_file and get_files
        """
        pass

class TestFileCheck:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133


    def test_ctor(self):
        match FileCheck():
            case API.AdaptiveMiddleware_p():
                assert(True)
            case x:
                assert(False), x

    def test_orphan_check(self, mocker, tmpdir):
        tmpfile = pl.Path(tmpdir) / "test.pdf"
        tmpfile.touch()
        assert(tmpfile.exists())
        field = model.Field("file", tmpfile)
        entry = model.Entry("test", "test:blah", [field])
        lib   = Library([entry])
        mid   = FileCheck()
        mid.pdf_is_modifiable = mocker.Mock(return_value=True)
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(MAPI.ORPHANED_K not in l2.entries[0].fields_dict)
                assert(not bool(lib.failed_blocks))
            case x:
                 assert(False), x


    def test_orphan_check_fail(self, tmpdir):
        tmpfile = pl.Path(tmpdir) / "test.pdf"
        assert(not tmpfile.exists())
        field = model.Field("file", tmpfile)
        entry = model.Entry("test", "test:blah", [field])
        lib   = Library([entry])
        mid   = FileCheck()
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(MAPI.ORPHANED_K in l2.entries[0].fields_dict)
                assert(not bool(lib.failed_blocks))
            case x:
                 assert(False), x


    def test_lock_check_fail(self, mocker, tmpdir):
        tmpfile = pl.Path(tmpdir) / "test.pdf"
        tmpfile.touch()
        assert(tmpfile.exists())
        field = model.Field("file", tmpfile)
        entry = model.Entry("test", "test:blah", [field])
        lib   = Library([entry])
        mid   = FileCheck()
        mid.pdf_is_modifiable = mocker.Mock(return_value=False)
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(MAPI.PDF_LOCKED_K in lib.entries[0].fields_dict)
                assert(not bool(lib.failed_blocks))
            case x:
                 assert(False), x


    @pytest.mark.skip
    def test_todo(self):
        """
        still to test:
        - epub existence
        """
        pass



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
import isbn_hyphenate
import pyisbn
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware, LibraryMiddleware)
from jgdv import Mixin, Proto

# ##-- end 3rd party imports

import bibble._interface as API
from . import _interface as MAPI
from bibble.util.middlecore import IdenBlockMiddleware
from bibble.util.mixins import ErrorRaiser_m

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--|

@Proto(API.ReadTime_p)
@Mixin(ErrorRaiser_m)
class IsbnValidator(IdenBlockMiddleware):
    """
      Try to validate the entry's isbn number
    """

    def on_read(self):
        Never()

    def transform_Entry(self, entry, library):
        match entry.get(MAPI.ISBN_K):
            case None:
                return [entry]
            case model.Field(value=str() as val) if bool(val):
                try:
                    isbn = pyisbn.Isbn(MAPI.ISBN_STRIP_RE.sub("", val))
                    if not isbn.validate():
                        raise pyisbn.IsbnError("validation fail")
                    else:
                        return [entry]
                except pyisbn.IsbnError as err:
                    # TODO add a fail block
                    self._logger.warning("ISBN validation fail: %s : %s", entry.key, val)
                    entry.set_field(model.Field(MAPI.INVALID_ISBN_K, val))
                    entry.set_field(model.Field(MAPI.ISBN_K, ""))
                    return [self.make_error_block(entry, err)]
            case model.Field(value=str() as val):
                del entry[MAPI.ISBN_K]
                return [entry]
            case x:
                raise TypeError(type(x))

